@@ -13,6 +13,7 @@
 # limitations under the License.
 
 import hashlib
+import json
 import shutil
 import os
 import sys
@@ -180,15 +181,10 @@ def _is_debugger_attached() -> bool:
     return debugger.is_attached()
 
 
-def _build_artifact_test_folder(
-    pytestconfig: Any, request: pytest.FixtureRequest, folder_or_file_name: str
-) -> str:
-    output_dir = pytestconfig.getoption("--output")
-    return os.path.join(
-        output_dir,
-        _truncate_file_name(slugify(request.node.nodeid)),
-        _truncate_file_name(folder_or_file_name),
-    )
+@pytest.fixture
+def output_path(pytestconfig: Any, request: pytest.FixtureRequest) -> str:
+    output_dir = Path(pytestconfig.getoption("--output")).absolute()
+    return os.path.join(output_dir, _truncate_file_name(slugify(request.node.nodeid)))
 
 
 def _truncate_file_name(file_name: str) -> str:
@@ -222,12 +218,13 @@ def browser_context_args(
 @pytest.fixture()
 def _artifacts_recorder(
     request: pytest.FixtureRequest,
+    output_path: str,
     playwright: Playwright,
     pytestconfig: Any,
     _pw_artifacts_folder: tempfile.TemporaryDirectory,
 ) -> Generator["ArtifactsRecorder", None, None]:
     artifacts_recorder = ArtifactsRecorder(
-        pytestconfig, request, playwright, _pw_artifacts_folder
+        pytestconfig, request, output_path, playwright, _pw_artifacts_folder
     )
     yield artifacts_recorder
     # If request.node is missing rep_call, then some error happened during execution
@@ -248,14 +245,33 @@ def browser_type(playwright: Playwright, browser_name: str) -> BrowserType:
     return getattr(playwright, browser_name)
 
 
+@pytest.fixture(scope="session")
+def connect_options() -> Optional[Dict]:
+    return None
+
+
 @pytest.fixture(scope="session")
 def launch_browser(
     browser_type_launch_args: Dict,
     browser_type: BrowserType,
+    connect_options: Optional[Dict],
 ) -> Callable[..., Browser]:
     def launch(**kwargs: Dict) -> Browser:
         launch_options = {**browser_type_launch_args, **kwargs}
-        browser = browser_type.launch(**launch_options)
+        if connect_options:
+            browser = browser_type.connect(
+                **(
+                    {
+                        **connect_options,
+                        "headers": {
+                            "x-playwright-launch-options": json.dumps(launch_options),
+                            **(connect_options.get("headers") or {}),
+                        },
+                    }
+                )
+            )
+        else:
+            browser = browser_type.launch(**launch_options)
         return browser
 
     return launch
@@ -307,8 +323,7 @@ class CreateContextCallback(Protocol):
         record_har_url_filter: Optional[Union[str, Pattern[str]]] = None,
         record_har_mode: Optional[Literal["full", "minimal"]] = None,
         record_har_content: Optional[Literal["attach", "embed", "omit"]] = None,
-    ) -> BrowserContext:
-        ...
+    ) -> BrowserContext: ...
 
 
 @pytest.fixture
@@ -462,12 +477,14 @@ class ArtifactsRecorder:
         self,
         pytestconfig: Any,
         request: pytest.FixtureRequest,
+        output_path: str,
         playwright: Playwright,
         pw_artifacts_folder: tempfile.TemporaryDirectory,
     ) -> None:
         self._request = request
         self._pytestconfig = pytestconfig
         self._playwright = playwright
+        self._output_path = output_path
         self._pw_artifacts_folder = pw_artifacts_folder
 
         self._all_pages: List[Page] = []
@@ -476,6 +493,12 @@ class ArtifactsRecorder:
         self._tracing_option = pytestconfig.getoption("--tracing")
         self._capture_trace = self._tracing_option in ["on", "retain-on-failure"]
 
+    def _build_artifact_test_folder(self, folder_or_file_name: str) -> str:
+        return os.path.join(
+            self._output_path,
+            _truncate_file_name(folder_or_file_name),
+        )
+
     def did_finish_test(self, failed: bool) -> None:
         screenshot_option = self._pytestconfig.getoption("--screenshot")
         capture_screenshot = screenshot_option == "on" or (
@@ -484,10 +507,8 @@ class ArtifactsRecorder:
         if capture_screenshot:
             for index, screenshot in enumerate(self._screenshots):
                 human_readable_status = "failed" if failed else "finished"
-                screenshot_path = _build_artifact_test_folder(
-                    self._pytestconfig,
-                    self._request,
-                    f"test-{human_readable_status}-{index+1}.png",
+                screenshot_path = self._build_artifact_test_folder(
+                    f"test-{human_readable_status}-{index + 1}.png",
                 )
                 os.makedirs(os.path.dirname(screenshot_path), exist_ok=True)
                 shutil.move(screenshot, screenshot_path)
@@ -500,11 +521,9 @@ class ArtifactsRecorder:
         ):
             for index, trace in enumerate(self._traces):
                 trace_file_name = (
-                    "trace.zip" if len(self._traces) == 1 else f"trace-{index+1}.zip"
-                )
-                trace_path = _build_artifact_test_folder(
-                    self._pytestconfig, self._request, trace_file_name
+                    "trace.zip" if len(self._traces) == 1 else f"trace-{index + 1}.zip"
                 )
+                trace_path = self._build_artifact_test_folder(trace_file_name)
                 os.makedirs(os.path.dirname(trace_path), exist_ok=True)
                 shutil.move(trace, trace_path)
         else:
@@ -524,12 +543,10 @@ class ArtifactsRecorder:
                     video_file_name = (
                         "video.webm"
                         if len(self._all_pages) == 1
-                        else f"video-{index+1}.webm"
+                        else f"video-{index + 1}.webm"
                     )
                     video.save_as(
-                        path=_build_artifact_test_folder(
-                            self._pytestconfig, self._request, video_file_name
-                        )
+                        path=self._build_artifact_test_folder(video_file_name)
                     )
                 except Error:
                     # Silent catch empty videos.
@@ -539,7 +556,8 @@ class ArtifactsRecorder:
                 # Can be changed to "if page.video" without try/except once https://github.com/microsoft/playwright-python/pull/2410 is released and widely adopted.
                 if video_option in ["on", "retain-on-failure"]:
                     try:
-                        page.video.delete()
+                        if page.video:
+                            page.video.delete()
                     except Error:
                         pass
 
